#include <iostream>
#include <string>
using std::cout ;
using std::cerr ;
using std::endl ;

using std::string ;

// Number of symbols in alphabet ;
const int symsize = 2 ;
// Alphabet
char alphabet[symsize] = {'0','1'} ;


#define EXERCISE_3_2_1
// Transition function for the state machine under consideration 
#ifdef EXERCISE_3_2_1
// Transition function for the DFA in exercise 3.2.1 on page 106
int trans[][symsize] = {
  {0,0},  // Unused slot, states start numbering from 1
  {2,1},
  {3,1},
  {3,2}
};
#endif
#ifdef EXERCISE_3_2_2
// Transition Function for the DFA in exercise 3.2.2 on Page 106
int trans[][symsize] = {
  {0,0}, // Unused slot, states start numbering from 1
  {2,3},
  {1,3},
  {2,1}
} ;
#endif



const int num_states = sizeof(trans)/sizeof(int)/symsize ;


// Regular Expression Operations

enum rops {EMPTY_SET, EMPTY_STRING, SYMBOL, CLOSURE, CONCAT, UNION} ;

class symbol {
public:
  char sym ;
  symbol(char c) { sym = c ; }
} ;


// class for regular expression tree
class regexp {
public:
  regexp *left, *right ;
  rops op_type ;
  char sym ;
  regexp() { op_type = EMPTY_SET ; left = 0 ; right = 0 ; }
  regexp(const symbol &s) { op_type = SYMBOL ; sym = s.sym ; left = 0 ;
  right = 0 ; }
  regexp(const enum rops &op) { op_type = op ; left = 0 ; right = 0 ; }
  operator string() ; 
} ;


// Convert regular expression tree representation to string
regexp::operator string()
{
  string sl,sr ;
  switch(op_type) {
  case EMPTY_SET:
    return string("Empty_Set") ;
  case EMPTY_STRING:
    return string("Empty_String") ;
  case SYMBOL:
    sl = "" ;
    sl += sym ;
    return sl ;
  case CLOSURE:
    if(left->op_type == SYMBOL) {
      string s = string(*left) + "*" ;
      return s ;
    } else {
      string s = string("(") + string(*left) + ")*" ;
      return s ;
    }
  case CONCAT:
    sl = string(*left) ;
    sr = string(*right) ;
    if(left->op_type == UNION)
      sl = string("(") + sl + string(")") ;
    if(right->op_type == UNION)
      sr = string("(") + sr + string(")") ;
    return sl + sr ;
  case UNION:
    return string(*left) + string(" + ") + string(*right) ;
  default:
    break ;
  }
  return string("ERROR") ;
}

regexp EmptySet(EMPTY_SET) ;
regexp EmptyString(EMPTY_STRING) ;


// Concatenation of two regexps
regexp *rconcat(regexp *s1,regexp *s2)
{
  // Anything Concatenated with the EMPTY_SET is the EMPTY_SET
  if(s1->op_type == EMPTY_SET)
    return s1 ;
  if(s2->op_type == EMPTY_SET)
    return s2 ;
  // The EMPTY_STRING is ignored by concatenation
  if(s1->op_type == EMPTY_STRING)
    return s2 ;
  if(s2->op_type == EMPTY_STRING)
    return s1 ;

    // Otherwise create concatenation operation
  regexp *p = new regexp ;
  p->op_type = CONCAT ;
  p->left = s1 ;
  p->right = s2 ;
  return p ;
}

// Kleene Closure
regexp *rclosure(regexp *s)
{
  // Closure(Closure) = Closure, Closure(EMPTY_STRING)=EMPTY_STRING
  if(s->op_type == CLOSURE || s->op_type == EMPTY_STRING)
    return s ;
  // Closure(EMPTY_SET) = EMPTY_STRING
  if(s->op_type == EMPTY_SET)
    return  &EmptyString ;
    
    // Else create closure operation
  regexp *p = new regexp(CLOSURE) ;
  p->left = s ;
  return p ;
}

// Union of two regular expressions
regexp *runion(regexp *s1, regexp *s2)
{
  // The Union of the EMPTY_SET is a null operation
  if(s1->op_type == EMPTY_SET)
    return s2 ;
  if(s2->op_type == EMPTY_SET)
    return s1 ;
    
    // The union of the EMPTY_STRING and closure is a null operation
  if(s1->op_type == EMPTY_STRING && s2->op_type == CLOSURE)
    return s2 ;
  if(s2->op_type == EMPTY_STRING && s1->op_type == CLOSURE)
    return s1 ;
  // If both sides are equal, the union is a null operation
  if(string(*s1) == string(*s2))
    return s1 ;
    
    // Create Union Operation
  regexp *p = new regexp(UNION) ;
  p->left = s1 ;
  p->right = s2 ;
  return p ;
}

// Delete any EMPTY_STRING from a list of union-ed expressions
// Used to remove epsilons from closure's
regexp *rdelete_epsilon(regexp *r)
{
  if(r->op_type == UNION) {
    if(r->left->op_type == EMPTY_STRING)
      return r->right ;
    if(r->right->op_type == EMPTY_STRING)
      return r->left ;
    r->left = rdelete_epsilon(r->left) ;
    r->right = rdelete_epsilon(r->right) ;
    return r ;
  }
  return r ;
}

// Optimize a regular expression  (remove redundant epsilons)
regexp *roptimize(regexp *r)
{
  switch(r->op_type) {
  case EMPTY_SET:
  case EMPTY_STRING:
  case SYMBOL:
    return r ;
  case CLOSURE:
    r->left = rdelete_epsilon(r->left) ;
    r->left = roptimize(r->left) ;
    return r ;
  case CONCAT:
    r->left = roptimize(r->left) ;
    r->right = roptimize(r->right) ;
    return r ;
  case UNION:
    r->left = roptimize(r->left) ;
    r->right = roptimize(r->right) ;
    return r ;
  }
  return r ;
}
    


// Return True if a regular expression contains the EMPTY_STRING
bool has_epsilon(regexp *r)
{
  if(r->op_type == EMPTY_STRING || r->op_type == CLOSURE)
    return true ;
  if(r->op_type == UNION) 
    return has_epsilon(r->left) || has_epsilon(r->right) ;
  return false ;
}

// Regular expressions for symbols
regexp *syms[symsize] ;

// Return a regular expression for the transition function (trans) that
// will take us from state i to state j passing through no state greater
// than k.
regexp *R(int i, int j, int k)
{
  if(k==0) {
    // Termination of the recursion.  (Rule 2.1 page 34)
    regexp *p = &EmptySet ;
    for(int m=0;m<symsize;m++)
      if(trans[i][m] == j)
        p = runion(p,syms[m]) ;
    if(i==j)
      p = runion(p,&EmptyString) ;
    return p ;
  }

  // Else compute R_ij^k = R_ik^k-1(R_kk^k-1)*R_kj^k-1 + R_ij^k-1
  // Optimize for cases where i==k and j==k
  regexp *rf ;
    
  if(i==k && j==k) {
    regexp *r1 = R(k,k,k-1) ;
    if(has_epsilon(r1)==true)
      rf = rclosure(r1) ;
    else 
      rf = rconcat(r1,rclosure(r1)) ;
  } else if(i==k) {
    regexp *r1 = R(k,k,k-1) ;
    regexp *r2 = R(k,j,k-1) ;
    rf = rconcat(rclosure(r1),r2) ;
  } else if(j==k) {
    regexp *r1 = R(i,k,k-1) ;
    regexp *r2 = R(k,k,k-1) ;
    rf = rconcat(r1,rclosure(r2)) ;
  } else {
    regexp *r1 = R(i,k,k-1) ;
    regexp *r2 = R(k,k,k-1) ;
    regexp *r3 = R(k,j,k-1) ;
    regexp *r4 = R(i,j,k-1) ;
    
    rf = runion(rconcat(r1,rconcat(rclosure(r2),r3)),r4) ;
  }
  return rf ;
}


main()
{
  // Define symbols represented in the transition function (trans)
  for(int i=0;i<symsize;++i)
    syms[i] = new regexp(symbol(alphabet[i])) ;

  for(int k=0;k<num_states;++k) {
    for(int i=1;i<num_states;++i)
      for(int j=1;j<num_states;++j)
        cout << "R_"<<i<<j<<"^"<<k<<"="<< string(*roptimize(R(i,j,k))) << endl ;
    cout << endl ;
  }
      
}
